suppressPackageStartupMessages(library(dplyr))
suppressPackageStartupMessages(library(purrr))
suppressPackageStartupMessages(library(stringr))
suppressPackageStartupMessages(library(readr))
suppressPackageStartupMessages(library(ggplot2))
suppressPackageStartupMessages(library(data.table))
suppressPackageStartupMessages(library(glue))
suppressPackageStartupMessages(library(tidyr))
readr.show_progress=FALSE

glue("Preprocessing data, this might take a while...")

get_all_data <- function(path) {
    list.files(path, recursive=TRUE, full.names=TRUE) %>%
    map(., function(x) str_extract(x, ".*data/.*csv")) %>% unlist %>%
    discard(is.na) %>%
    map_df(read_csv, col_types=cols(Section = col_character()))
}

get_specific_data <- function(path, bench=NA, court=NA) {
    df <- get_all_data(path)
    if (!is.na(bench) & !is.na(court)) df %>% filter(Bench==bench, Court==court)
    else if (!is.na(bench) & is.na(court)) df %>% filter(Bench==bench)
    else df
}

df <- get_all_data(path=path)
df_split <- split(df, df$Bench) %>% map(., function(df) split(df, df$Bench))


## UNIQUE COUNTS

sno               <- unique(df$`S.No`)                           %>% length
cpca              <- unique(df$`CP/CA No`) %>% sort              %>% length
case_purpose      <- unique(df$`Case Purpose`)                   %>% length
section           <- unique(df$`Section`)                        %>% length
name_parties      <- unique(df$`Name of Parties`)                %>% length
name_practitioner <- unique(df$`Name of the Legal Practitioner`) %>% length
remarks           <- unique(df$`Remarks`)                        %>% length
coram             <- unique(df$`Coram`)                          %>% length

## PLOT num. hearings per bench-court

tmp <- df %>%
    group_by(Bench, Court) %>%
    summarise(num=n()) %>%
    unite(Bench_Court, c(Bench,Court)) %>%
    ### PLOT
    ggplot(aes(x = Bench_Court, y = num)) +
    theme_light() +
    geom_bar(position="stack", stat="identity",
             color="steelblue", fill="steelblue") +
    theme(axis.text.x = element_text(angle=90, hjust=1, face="bold")) +
    xlab("\nBench - Court\n") + ylab("\nFrequency\n") +
    ggtitle("\nFrequency of Hearings per Court in every Bench\n") +
    theme(plot.title = element_text(hjust = 0.5, face="bold"))

## Key to the fields

dtypes <- c("Numerical", "Character", "Factor",
            "Factor",    "Character", "Factor",
            "Factor",    "Character", "Character",
            "Factor",    "Factor")

uniques <- c()
for (col in colnames(df)) uniques = c(uniques, (unique(df[[col]]) %>% length))

descriptions <-
    c("Serial number of the hearing. Not always in sequential order",
      "Date of the hearing",
      "Name of the bench",
      "Courtroom number",
      "Unique case identifier",
      "Purpose of listing",
      "Section of the law the case falls under",
      "Names of the petitioner and respondent",
      "Names of the petitoner advocate and the respondent advocate (if applicable)",
      "Remarks",
      "Names of members of the coram")

get_egs <- function(df, col, n=10, sep=", ") str_c(sample(unique(df[[col]]),n),collapse=sep)
formats_subcats <-
    c("Eg: 1, 2, 3",
      "YYYY-MM-DD",
      glue('Eg: {get_egs(df, "Bench", 5)}'),
      glue('Eg: {get_egs(df, "Court", 5)}'),
      glue('Eg: {get_egs(df, "CP/CA No", 3)}'),
      glue('Eg: Hearing; For further consideration; New petition; Admission; For order; ',
           get_egs(df, "Case Purpose", 3, "; ")),
      glue('Eg: 252; 7 IBC; U/s 9 of IBC Code,2016; ', get_egs(df, "Section", 2, "; ")),
      glue('Eg: ', get_egs(df, "Name of Parties", 2, "\n")),
      glue('Eg: ', get_egs(df, "Name of the Legal Practitioner", 2, "\n")),
      glue('Eg: NA, Allowed, Disposed, ', get_egs(df, "Remarks", 3)),
      glue('Eg: ', get_egs(df, "Coram", 2, "; ")))

colms <- colnames(df)
df_summary <-
    #tibble(colms, descriptions, dtypes, formats_subcats, uniques) %>%
    tibble(colms, dtypes, formats_subcats, uniques) %>%
    rename(Field       = colms,
           #Description = descriptions,
           `Data Type` = dtypes,
           `Uniques` = uniques,
           `Formats or Sub-categories*` = formats_subcats)

## Frequency Distribution of Fields

col_freq_one <- function(df, colname) {
    group_by(df, !!sym(colname)) %>%
        summarise(Frequency=n()) %>%
        arrange  (desc(Frequency)) %>%
        mutate   (`%` = 100 * Frequency / sum(Frequency))
}

col_freq_all <- function(df) {
    freqs        <- map(colnames(df), function(col) col_freq_one(df, col))
    names(freqs) <- colnames(df)
    freqs
}

lowercase_df <- function(df, prefix="", suffix="",
                         cols=c("Case Purpose", "Remarks", "Section")) {
    for(col in cols)
        df = mutate(df, !!sym(glue(prefix, col, suffix)) := tolower(!!sym(col)))
    df
}

freq_fields_all       <- col_freq_all(df)
freq_fields_all_lower <- col_freq_all(lowercase_df(df))

ffa  <- freq_fields_all
ffal <- freq_fields_all_lower

gather_remarks_dates <- function(df) {
    mutate(df, Remarks = ifelse(
                   str_detect(Remarks,"^\\d{1,4}.+\\d{1,2}.+\\d{1,4}$"),
                              "date (YYYY-MM-DD), (DD-MM-YYYY), etc.",
                              Remarks))
}

## after removing dates from remarks
ffa_dt  <- gather_remarks_dates(df) %>% col_freq_all
ffal_dt <- gather_remarks_dates(lowercase_df(df)) %>% col_freq_all

ffa_nodt  <-
    df %>%
    filter(!str_detect(Remarks, "^\\d{1,4}.+\\d{1,2}.+\\d{1,4}$")) %>%
    col_freq_all
ffal_nodt <-
    lowercase_df(df) %>%
    filter(!str_detect(Remarks, "^\\d{1,4}.+\\d{1,2}.+\\d{1,4}$")) %>%
    col_freq_all

## ffa_dt$Remarks
## ffal_dt$Remarks
## ffa_nodt$Remarks
## ffal_nodt$Remarks

## map(list(ffa_dt, ffal_dt, ffa_nodt, ffal_nodt), function(df) df$Remarks)

freq_bc <-
    df %>%
    unite(bc, c(Bench, Court), sep=": ") %>%
    col_freq_one("bc")

#######

filter_string <- function(df_list, colname, string) {
    df_list[[colname]] %>%
        select(-`%`) %>%
        filter(str_detect(!!sym(colname), string)) %>%
        data.frame
}

combine_filter_string <- function(df, colname, str1, str2) {
    bind_rows(df %>% filter_string(colname, str1),
              df %>% filter_string(colname, str2)) %>%
    arrange(desc(Frequency))
}


### EXPLORING

num_unique_cpcas <- unique((lowercase_df(df))$`CP/CA No`) %>% length

cpca_recurring <-
    lowercase_df(df) %>%
    col_freq_one("CP/CA No") %>%
    filter(Frequency > 2)  %>% .$`CP/CA No`

cpca_splits <-
    lowercase_df(df) %>% filter(`CP/CA No` %in% cpca_recurring) %>%
    split(., .$`CP/CA No`)


#############################################################
#############################################################
####### ALL INCLUSIVE FUNCTIONS FOR PRODUCTION SCRIPT #######

get_all_data <- function(path="~/Desktop/NCLT") {
    list.files(path, recursive=TRUE, full.names=TRUE) %>%
    map(., function(x) str_extract(x, ".*data/.*csv")) %>% unlist %>%
    discard(is.na) %>%
    map_df(read_csv, col_types=cols(Section = col_character()))
}

lowercase_df <- function(df, prefix="", suffix="",
                         cols=c("Case Purpose", "Remarks", "Section")) {
    for(col in cols)
        df = mutate(df, !!sym(glue(prefix, col, suffix)) := tolower(!!sym(col)))
    df
}

gen_regex <- function(keywords) {
    pats <- "("
    ## iterate from first to n-1 elements
    idxs <- seq_along(keywords)[1:length(keywords)-1]
    for (i in idxs) {
        pats = glue(pats, keywords[i], "|")
    }
    ## add nth/last element
    glue(pats, keywords[length(keywords)], ")")
}

add_case_status <- function(DF,
                            birth_keywords = c("new petition", "admission"),
                            death_keywords = c("dispose", "dismis",
                                               "order pronounced", "allow")
                            ) {
    birth_regex <- gen_regex(birth_keywords)
    death_regex <- gen_regex(death_keywords)
    lowercase_df(DF) %>%
    split(., .$`CP/CA No`) %>%
    map_df(., function(df) {
        birth  <- filter(df, str_detect(`Case Purpose`, birth_regex))
        death  <- filter(df, str_detect(Remarks,        death_regex))
        if ((nrow(birth) > 0) && (nrow(death) > 0)) {
            df = mutate(df, category = "FULL")
            df
        } else if ((nrow(birth) > 0) && (nrow(death) == 0)) {
            df = mutate(df, category = "BIRTH ONLY")
            df
        } else if ((nrow(birth) == 0) && (nrow(death) > 0)) {
            df = mutate(df, category = "DEATH ONLY")
            df
        } else if ((nrow(birth) == 0) && (nrow(death) == 0)){
            df = mutate(df, category = "ONGOING")
            df
        }
    })
}

## path <- "~/Desktop/other/IGIDR/NCLT/"
## df   <- get_all_data(path=path)

birth_keywords <- c("new petition", "admission")
death_keywords <- c("dispose", "dismis", "allow", "order pronounced")
case_statuses  <- add_case_status(df, birth_keywords, death_keywords)



########### END OF ALL INCLUSIVE PRODUCTION SCRIPT ############
###############################################################
###############################################################



cases_ongoing    <- case_statuses %>% filter(category=="ONGOING")
cases_birth_only <- case_statuses %>% filter(category=="BIRTH ONLY")
cases_death_only <- case_statuses %>% filter(category=="DEATH ONLY")
cases_full       <- case_statuses %>% filter(category == "FULL")

ongoing_cpcas <- unique(cases_ongoing$`CP/CA No`)
birth_cpcas   <- unique(cases_birth_only$`CP/CA No`)
death_cpcas   <- unique(cases_death_only$`CP/CA No`)
full_cpcas    <- unique(cases_full$`CP/CA No`)

### output
cases_full_raw <- df %>% filter(`CP/CA No` %in% full_cpcas)
### save to disk

cpca_splits <-
    lowercase_df(df) %>% filter(`CP/CA No` %in% cpca_recurring) %>%
    split(., .$`CP/CA No`)

full_cases_split <-
    case_statuses %>% filter(`CP/CA No` %in% full_cpcas) %>%
    split(., .$`CP/CA No`)

cats <- c("ONGOING", "BIRTH ONLY", "DEATH ONLY", "FULL")
cpcas <- c(ongoing_cpcas %>% length, birth_cpcas %>% length,
           death_cpcas %>% length, full_cpcas %>% length)

category_distribution <-
    tibble(
        Category = cats,
        `# Unique CP/CAs` = cpcas
    ) %>% mutate(cum_freq = cumsum(`# Unique CP/CAs`),
                 `%` = `# Unique CP/CAs`*100 / sum(`# Unique CP/CAs`))



select_few <- function(df) select(df, Date, `CP/CA No`, `Case Purpose`, Remarks)


time_distribution <-
    map_df(full_cases_split, function(df) {
        num_hearings <- nrow(df)
        dates <- df$Date
        timespan <- max(dates) - min(dates)
        bench <- unique(df$Bench)
        tibble(`CP/CA No.` = unique(df$`CP/CA No`),
               `Bench` = bench,
               `No. of Hearings` = num_hearings,
               `Timespan` = timespan) %>% tbl_df
    })

perc_timespan <- function(df,n)
    ((filter(df,Timespan < n) %>% nrow) * 100 / nrow(df)) %>% round(2)


num_cps <- names(full_cases_split) %>% str_detect("^[Cc].?[Pp]") %>% sum

num_cas <- names(full_cases_split) %>% str_match("^\\(?[Cc].?[Aa]") %>%
    str_detect("CAA", negate=TRUE) %>% sum(na.rm=TRUE)

section_names <- map_chr(full_cases_split, function(df) unique(df$Section)[1])
section_names_df <-
    data.frame(`CP/CA No` = names(section_names),
               Section = section_names, stringsAsFactors=FALSE) %>%
    tbl_df

section_freq <-
    section_names_df %>%
    group_by(Section) %>%
    summarise(freq_section = n()) %>%
    arrange(desc(freq_section)) %>%
    mutate(`%` = freq_section*100/sum(freq_section))

section_freq_condensed_mild <-
    section_freq %>%
    mutate(Section_ = case_when(str_detect(Section, "7.*ibc") ~ "IBC - Section 7",
                                str_detect(Section, "9.*ibc") ~ "IBC - Section 9",
                                str_detect(Section, "\\b25\\d{1}\\b") ~ "Companies Act (25_)",
                                str_detect(Section, "\\b23\\d{1}\\b") ~ "Companies Act (23_)")) %>%
    group_by(Section_) %>%
    summarise(freq = sum(freq_section)) %>%
    mutate(`%` = freq*100/sum(freq)) %>%
    arrange(desc(`%`)) %>%
    mutate(cumfreq = cumsum(freq)) %>%
    select(Section_, freq, cumfreq, `%`)

glue("... finished preprocessing data.")
