suppressPackageStartupMessages(library(dplyr))
suppressPackageStartupMessages(library(glue))
suppressPackageStartupMessages(library(purrr))
suppressPackageStartupMessages(library(stringr))
suppressPackageStartupMessages(library(readr))
library(optparse)

option_list <- list(
    make_option(c('-i', '--path_data'), type='character',
                default='~/Desktop/other/IGIDR/NCLT/',
                help='Root path where all the NCLT data is stored'),
    make_option(c('-o', '--path_output'), type='character',
                default=getwd(),
                help='Path where the output must be saved. Default = current directory.'),
    make_option(c('-f', '--output_filename'), type='character',
                default='01_cpca_numbers',
                help='Name of the output file with all the CP/CA Nos. of cases
                whose entire lifespan exists in the dataset
                Default: "01_cpca_numbers"'),
    make_option(c('-b', '--birth_keywords'), type='character',
                default='new petition;admission',
                help='Keywords to recognise the birth of a case (from the "Case Purpose" column)
                Should be enclosed in quotes. If entering multiple, they must be
                separated by a ";".
                Default: "new petition;admission"'),
    make_option(c('-d', '--death_keywords'), type='character',
                default='dispose;dismis;order pronounced;allow',
                help='Keywords to recognise the death of a case (from the "Remarks" column).
                Should be enclosed in quotes.
                If entering multiple, they must be separated by a ";"
                Default="dispose;dismis;order pronounced;allow"'),
    make_option(c('-t', '--case_types'), type='character',
                default='FULL;BIRTH-ONLY;DEATH-ONLY;ONGOING',
                help='Filters for the 4 case-types. If entering multiple, separate with ";"
                and enclose in quotes.
                "FULL" => Cases whose entire lifespan is within the dataset
                "BIRTH-ONLY" => Cases that were born in the dataset but have not died yet
                "DEATH-ONLY" => Cases that were born before but died in the dataset
                "ONGOING" => Cases that were born before and are still ongoing
                Default: "FULL;BIRTH-ONLY;DEATH-ONLY;ONGOING"')
)

get_all_data <- function(path) {
    list.files(path, recursive=TRUE, full.names=TRUE) %>%
    map(., function(x) str_extract(x, ".*data/.*csv")) %>% unlist %>%
    discard(is.na) %>%
    map_df(read_csv, col_types=cols(Section = col_character()))
}

lowercase_df <- function(df, prefix="", suffix="",
                         cols=c("Case Purpose", "Remarks", "Section")) {
    for(col in cols)
        df = mutate(df, !!sym(glue(prefix, col, suffix)) := tolower(!!sym(col)))
    df
}

gen_regex <- function(keywords) {
    pats <- "("
    ## iterate from first to n-1 elements
    idxs <- seq_along(keywords)[1:length(keywords)-1]
    for (i in idxs) {
        pats = glue(pats, keywords[i], "|")
    }
    ## add nth/last element
    glue(pats, keywords[length(keywords)], ")")
}

case_status_categories <- list(
    full                    = "FULL",
    born_not_dead           = "BIRTH-ONLY",
    born_before_died_here   = "DEATH-ONLY",
    born_before_still_alive = "ONGOING"
)

add_case_status <- function(DF,
                            birth_keywords = c("new petition", "admission"),
                            death_keywords = c("dispose", "dismis",
                                               "order pronounced", "allow"),
                            case_status_categories = case_status_categories
                            ) {
    birth_regex <- gen_regex(birth_keywords)
    death_regex <- gen_regex(death_keywords)
    lowercase_df(DF) %>%
    split(., .$`CP/CA No`) %>%
    map_df(., function(df) {
        birth  <- filter(df, str_detect(`Case Purpose`, birth_regex))
        death  <- filter(df, str_detect(Remarks,        death_regex))
        if ((nrow(birth) > 0) && (nrow(death) > 0)) {
            df = mutate(df, category = case_status_categories$full)
            df
        } else if ((nrow(birth) > 0) && (nrow(death) == 0)) {
            df = mutate(df, category = case_status_categories$born_not_dead)
            df
        } else if ((nrow(birth) == 0) && (nrow(death) > 0)) {
            df = mutate(df, category = case_status_categories$born_before_died_here)
            df
        } else if ((nrow(birth) == 0) && (nrow(death) == 0)){
            df = mutate(df,
                        category = case_status_categories$born_before_still_alive)
            df
        }
    })
}

program_summary <- "This script takes as input the path to the scraped NCLT data, and a list
of keywords each to determine death and birth of a case, and outputs a text file
containing all the CP/CA numbers of cases that were born in and died within the
timeframe of this dataset. Each CP/CA number is stored in a newline, separated by '\\n'.


Usage Example
=============

Rscript gen_full_cpcas.R \
    --path_data       ~/NCLT-SCRAPED-DATA \
    --path_output     ~/ \
    --output_filename full_cpcas \
    --birth_keywords  'new petition;admission' \
    --death_keywords  'dispose;dismis;order pronounced;allow' \
    --case_types      'FULL;BIRTH-ONLY'
"

opt <- parse_args(OptionParser(option_list=option_list, usage=program_summary))

path_data  <- opt$path_data
path_out   <- opt$path_output
out_fname  <- opt$output_filename
case_types <- opt$case_types
message("Parsing Arguments")
birth_keywords <- opt$birth_keywords %>% strsplit(';') %>% .[[1]] %>% gen_regex
death_keywords <- opt$death_keywords %>% strsplit(';') %>% .[[1]] %>% gen_regex

df <- get_all_data(path_data)
glue("Read data")

glue("Classifying case categories...")
case_statuses <- add_case_status(df, birth_keywords,
                                 death_keywords, case_status_categories)
glue("... Done")

glue(case_types)
cases     <- case_statuses %>% filter(category %in% str_split(case_types, ";")[[1]])
#cases_full     <- case_statuses %>% filter(category == case_status_categories$full)

cpcas     <- unique(cases$`CP/CA No`)

out_file <- glue(path_out, "/", out_fname, ".txt")
cat(cpcas, sep="\n", file=out_file)

glue("Saved {length(cpcas)} CPCA numbers of cases that were born in and died within the dataset to {out_file}")
