##
## Description: This scripte visits 'https://nclt.gov.in/exposed-status-of-cause-list-page'),
## grabs all the bench-court combinations available, and saves them to
## a file called `benches_courts_names.pkl`. Run `python _01_get_benches_courts.py -h` to see info on the arguments you can pass in.
##
## Author: Rahul Somani
## Date Last Revised: 06-June-2020
##

from imports import *
from utils import *

def get_benches(browser: Firefox, # after browser.get(url_to_causelist_page)
                id_bench:str = 'edit-field-bench-target-id'
               ) -> list:
    'Get list of benches from "Customised Cause List" homepage'
    soup = BeautifulSoup(browser.page_source, features='lxml')
    benches = soup.find(id=id_bench).find_all('option')
    return [bench.contents[0] for bench in benches if not bench.contents[0]=='Select One']

def get_courts(browser:Firefox,
               benches:list,
               id_bench:str = 'edit-field-bench-target-id',
               id_court:str = 'edit-field-bench-court-target-id-entityreference-filter',
               sleep:Union[int, float] = 2.0
              ) -> dict:
    bench_courts = {}
    'For a list of benches, find the courts that fall under the bench'
    for bench in benches:
        select_bench = Select(browser.find_element_by_id(id_bench))
        select_bench.select_by_visible_text(bench)
        time.sleep(sleep) # let the page load the court options

        soup = BeautifulSoup(browser.page_source, features='lxml')
        courts = soup.find(id=id_court).find_all('option')
        courts = [court.contents[0] for court in courts if not court.contents[0]=='- Any -']
        bench_courts[bench] = courts
        print(f'------------------ {bench} ------------------')
        print(courts)
        print(); print('============================================================')
    return bench_courts

def dict2df(bench_court_dict:dict) -> pd.DataFrame:
    bench_, court_ = [],[]
    for bench,courts in bench_court_dict.items():
        for court in courts:
            bench_.append(bench)
            court_.append(court)
    return pd.DataFrame({
        'bench': bench_,
        'court': court_
    })

def get_bench_courts(url:str='https://nclt.gov.in/exposed-status-of-cause-list-page',
                     #url:str = 'https://nclt.gov.in/exposed-customised-cause-list-page',
                     headless_browser:bool = True,
                     return_browser:bool = False,
                     id_bench:str = 'edit-field-bench-target-id',
                     id_court:str = 'edit-field-bench-court-target-id-entityreference-filter',
                     sleep:float = 2.0,
                     verbose:bool = True) -> Union[dict, Tuple[dict, Firefox]]:
    '''
    Grab all unique bench-court combos from the "Exposed/Customised Cause List" webpage.
    If `return_browser` is True, then a tuple of (bench_courts:dict, browser) is returned,
    else a dict of bench-court combos is returned.
    '''
    # Setup browser
    if verbose: print('Setting up the virtual browser')
    options = FirefoxOptions()
    options.headless = True if headless_browser else False
    browser = Firefox(options=options)
    browser.get(url)
    if verbose: print('Opened the cause lists homepage')

    # get benches, courts per bench
    if verbose: print('Grabbing benches')
    benches      = get_benches(browser, id_bench)
    if verbose: print(f'Grabbed benches ==> {benches}')
    if verbose: print()
    if verbose: print('Grabbing courts per bench')
    bench_courts = get_courts(browser, benches, id_bench, id_court, sleep)

    # check if court grabbing process had any errors
    flatten = lambda l: [item for sublist in l for item in sublist]
    courts  = flatten([court for bench,court in bench_courts.items()])
    if verbose: print('Checking for potential errors')
    while 'Select One' in courts: # error checking
        if verbose: print('Previous pass of grabbing courts per bench was unsuccessful. Retrying')
        bench_courts = get_courts(browser, benches, id_bench, id_court, sleep)
        courts = flatten([court for bench,court in bench_courts.items()])
    if verbose: print(f'Grabbed unique court-bench combinations: {bench_courts}')

    if return_browser: return bench_courts, browser
    else:
        browser.quit()
        return bench_courts

@call_parse
def main(url:Param('URL of the NCLT Cause Lists (HTML) Webpage',
         str) = 'https://nclt.gov.in/exposed-status-of-cause-list-page',
         headless:Param('Run browser in headless mode?', bool_arg) = True,
         id_bench:Param('HTML `id` of the bench entry', str) = 'edit-field-bench-target-id',
         id_court:Param('HTML `id` of the court entry', str) = 'edit-field-bench-court-target-id-entityreference-filter',
         sleep_before_click:Param('How long to wait, in seconds, before clicking the `apply` button', float) = 2.0,
         verbose:Param('Verbose output?', bool_arg) = True,
         save_path:Param('Path to save the `.pkl` file (dict) to', str) = os.getcwd(),
         save_fname:Param('Name of saved file', str) = 'benches_courts_names.pkl'):

         bench_courts = get_bench_courts(url=url, headless_browser=headless,
                                         return_browser=False,
                                         id_bench=id_bench, id_court=id_court,
                                         sleep=sleep_before_click, verbose=verbose)

         if save_path is not None:
             save_path = Path(save_path)/save_fname
             save_pickle(any_variable=bench_courts, save_path=save_path)
             print(f'Saved `.pkl` file to {save_path/save_fname}.')
             print(f'Run `python verify_benches_courts.py` to automatically check if these are valid.')
         print(dict2df(bench_courts))
