## Description:Takes as input the verified
##   `benches_courts-VALIDATED.pkl` file from the 2nd script and grabs
##   sample URLs for those bench-court combinations and saves the to 
##   `bench_court_urls.csv`. If some URLs are skipped (a warning message is
##   displayed saying so), they are saved to `bench_court_skipped.csv` . You
##   can re-run this script until all URLs have been grabbed.
## Run `python _03_get_bench_court_urls.py -h` to see info on the arguments you can pass in.
## Author: Rahul Somani
## Date Last Revisited: 06-June-2020


from utils import *
from imports import *
from html_parsing import *
from _01_get_benches_courts import *

def validate_and_get_urls(bench_courts:dict,
                          browser:Firefox=0,
                          browser_headless:bool=True,
                          url:str = 'https://nclt.gov.in/exposed-status-of-cause-list-page',
                          verbose:bool = True,
                          sleep_select:float = 2.0,
                          sleep_pageload:float = 5.0,
                          wait_after_load:float = 15.0,
                          validation_class_name:str = 'view-id-status_of_cause_list'
                          ) -> pd.DataFrame:

    # if browser doesn't exist, create one
    try: browser.current_url
    except AttributeError:
        options = FirefoxOptions()
        options.headless = browser_headless
        browser = Firefox(options=options)

    bench_, court_, urls = [],[],[]
    bench_skipped, court_skipped = [],[]
    target_id, entity_reference_id = [],[]

    for bench,courts in bench_courts.items():
        for court in courts:
            try:
                browser = open_cause_list(browser, bench,
                                          court, pd.Timestamp('20200103'),
                                          sleep_select=sleep_select,
                                          sleep_pageload=sleep_pageload,
                                          verbose=verbose)
            except NoSuchElementException:
                print('''
                Bench-court combinations were NoSuchElementExceptiond incorrectly.
                Either retry running the validation check, or
                check the bench-court combinations manually.
                ''')
                return
            # time.sleep(wait_after_load)
            # wait until the page url has loaded
            try:
                WebDriverWait(browser, wait_after_load).until(
                    EC.presence_of_element_located((By.CLASS_NAME, validation_class_name))
                )
                if verbose: print(f'Browser URL: {browser.current_url}')
                if verbose: print(f'==================================================================')
                if verbose: print();print()
                bench_.append(bench)
                court_.append(court)
                urls.append(browser.current_url)
            except TimeoutException:
                print(f'Page not loading. Skipping {bench} -- {court}')
                bench_skipped.append(bench)
                court_skipped.append(court)
                pass
    target_id = [(extract_target_id(url)) for url in urls]
    entity_reference_id = [(extract_entity_reference_id(url)) for url in urls]
    df = pd.DataFrame({'bench': bench_,
                        'court': court_,
                        'target_id': target_id,
                        'entity_reference_id': entity_reference_id,
                        'sample_url': urls})
    return (df, bench_skipped, court_skipped)

@call_parse
def main(path_pkl:Param('Path to the pickled file which is a dictionary of bench-court combinations',
         str) = 'benches_courts-VALIDATED.pkl',
         benches:Param('Grab URLs for these benches only', str, nargs='*')=None,
         save_path:Param('Path to save the dataframe of extracted URLs to', str) = os.getcwd(),
         save_fname:Param('Name of the saved dataframe', str) = 'bench_court_urls.csv',
         append:Param("Append to existing dataframe (DF). Use this if you're only grabbing urls for selected benches and want to add it to an existing DF. Ensure that the filename of the existing DF is the same as `save_fname`",
         bool_arg)=True,
         headless_browser:Param('Run browser in headless mode?', bool_arg)=True,
         url:Param('URL to the HTML cause lists homepage', str) = 'https://nclt.gov.in/exposed-status-of-cause-list-page',
         verbose:Param('Verbose output?', bool_arg) = True,
         sleep_select:Param('How long to sleep after entering the court', float) = 2.0,
         sleep_pageload:Param('How long to wait after clicking the `apply` button?', float) = 4.0,
         wait_after_load:Param('Max amount of time to wait for the cause list to load', float) = 15.0):
    bench_courts = load_pickle(path_pkl)
    save_path = Path(save_path)

    if benches is not None: getters = {bench: bench_courts.get(bench) for bench in benches}
    else: getters = bench_courts
    print(getters)
    df,skipped_bench,skipped_court = validate_and_get_urls(
        bench_courts=getters,
        browser_headless=headless_browser,
        url=url,
        verbose=verbose,
        sleep_select=sleep_select,
        sleep_pageload=sleep_pageload,
        wait_after_load=wait_after_load
        )
    if skipped_bench==[]: print('No benches were skipped. All URLs have been grabbed successfully')
    else:
        print('Some bench-court combinations were skipped:')
        skipped = pd.DataFrame({
            'SKIPPED BENCH': skipped_bench,
            'SKIPPED COURT': skipped_court
        })
        print(f'Saving skipped bench-court combinations to {save_path/"bench_court_skipped.csv"}')
        print(skipped)
        skipped.to_csv(save_path/"bench_court_skipped.csv", index=False)

    if append:
        if (save_path/save_fname).exists():
            print('Appending to existing DF')
            existing_df = pd.read_csv(save_path/save_fname)
            new_df = pd.concat([existing_df, df])
            new_df = new_df.sort_values(['bench', 'court'])
            new_df = new_df.drop_duplicates().reset_index(drop=True)
            new_df.to_csv(save_path/save_fname, index=False)
            # oddly, all duplicates aren't dropped so read it in again and send it
            new_df = pd.read_csv(save_path/save_fname)
            new_df = new_df.drop_duplicates().reset_index(drop=True)
            new_df = new_df.sort_values(['bench', 'court'])
            new_df.to_csv(save_path/save_fname, index=False)
        else:
            print(f"{save_path/save_fname} doesn't exist. Creating new DF")
            df.to_csv(save_path/save_fname, index=False)
    else:
        print('Succesfully grabbed the following bench-court combinations:')
        df.to_csv(save_path/save_fname, index=False)
